/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import java.util.List;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.ozgcloud.eingang.formcycle.FormCycleConfirmationResponse;
import de.ozgcloud.formcycle.attachment.AttachmentMapper;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.ozgcloud.formcycle.errorhandling.OzgPluginSoftError;
import de.ozgcloud.formcycle.errorhandling.Warning;
import de.ozgcloud.formcycle.formdata.FormData;
import de.ozgcloud.formcycle.formdata.OzgCloudFormDataMapper;
import de.ozgcloud.formcycle.formdata.PluginFormDataAdapter;
import de.xima.fc.entities.Attachment;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
public final class OzgPluginExecutor {

	private static final Logger LOG = LoggerFactory.getLogger(OzgPluginExecutor.class);

	private final OzgCloudFormDataMapper ozgCloudFormDataMapper;
	private final OzgCloudFormDataHttpClient ozgHttpClient;
	private final AttachmentMapper attachmentMapper;
	private final PluginFormDataAdapter pluginFormDataAdapter;
	private final Supplier<List<Attachment>> attachmentsSupplier;

	public ExecutionResult execute() {
		LOG.debug("Executing plugin WorkflowElementNodePlugin");
		var formData = addRepresentations(pluginFormDataAdapter.readFormData());
		formData = addAttachments(formData);
		return buildExecutionResult(sendFormData(formData), formData);
	}

	FormData addRepresentations(FormData formData) {
		var representations = getAttachedFiles(attachmentUuid -> !formData.getAttachmentUuids().contains(attachmentUuid));
		if (representations.isEmpty()) {
			return formData.toBuilder().warning(buildMissingRepresentationsWarning(formData.getRequestId())).build();
		}
		return formData.toBuilder().representations(representations).build();
	}

	Warning buildMissingRepresentationsWarning(String requestId) {
		return buildWarning(OzgPluginSoftError.MISSING_REPRESENTATION,
				String.format("Representation is missing. Ensure workflow is configured to attach form view to form data. (Request ID: %s)",
						requestId));
	}

	FormData addAttachments(FormData formData) {
		if (formData.getAttachmentUuids().isEmpty()) {
			return formData;
		}
		var attachments = getAttachedFiles(formData.getAttachmentUuids()::contains);
		if (attachments.isEmpty()) {
			return formData.toBuilder().warning(buildMissingAttachmentsWarning(formData.getRequestId())).build();
		}
		return formData.toBuilder().attachments(attachments).build();
	}

	List<FormcycleAttachment> getAttachedFiles(Predicate<String> attachmentUuidFilter) {
		return attachmentsSupplier.get().stream().filter(attachment -> attachmentUuidFilter.test(attachment.getUUID())).map(attachmentMapper::map)
				.collect(Collectors.toList());
	}

	Warning buildMissingAttachmentsWarning(String requestId) {
		return buildWarning(OzgPluginSoftError.MISSING_ATTACHMENTS,
				String.format("Attachments uploaded in the form were not included in the form data. (RequestId: %s)", requestId));
	}

	private Warning buildWarning(OzgPluginSoftError warningCode, String message) {
		return Warning.builder().errorCode(warningCode).message(message).build();
	}

	FormCycleConfirmationResponse sendFormData(FormData formData) {
		return ozgHttpClient.send(ozgCloudFormDataMapper.map(formData), formData.getAttachments(), formData.getRepresentations());
	}

	ExecutionResult buildExecutionResult(FormCycleConfirmationResponse response, FormData formData) {
		LOG.debug("Formcycle adapter response: {}", response);
		return ExecutionResult.builder().vorgangnummer(response.getVorgangNummer()).warnings(formData.getWarnings()).build();
	}
}
