/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.net.URL;
import java.util.Collections;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

class BuildPropertiesTest {

	@Spy
	@InjectMocks
	private BuildProperties buildProperties;

	@Mock
	private Manifest manifest;
	@Mock
	private Attributes attributes;

	@Nested
	class TestReadPluginVersion {

		private static final String PLUGIN_VERSION = "version";

		@Mock
		private URL manifestUrl;

		@Nested
		class TestSuccess {
			@BeforeEach
			void init() {
				when(manifest.getMainAttributes()).thenReturn(attributes);
				when(attributes.getValue(anyString())).thenReturn(PLUGIN_VERSION);

				doReturn(Collections.enumeration(List.of(manifestUrl))).when(buildProperties).loadManifests();
				doReturn(manifest).when(buildProperties).readManifest(any());
				doReturn(true).when(buildProperties).isOzgPluginManifest(any());
			}

			@Test
			void shouldCallLoadManifests() {
				buildProperties.readPluginVersion();

				verify(buildProperties).loadManifests();
			}

			@Test
			void shouldCallReadManifest() {
				buildProperties.readPluginVersion();

				verify(buildProperties).readManifest(manifestUrl);
			}

			@Test
			void shouldCallIsOzgPluginManifest() {
				buildProperties.readPluginVersion();

				verify(buildProperties).isOzgPluginManifest(manifest);
			}

			@Test
			void shouldReturnVersion() {
				var version = buildProperties.readPluginVersion();

				assertThat(version).isEqualTo(PLUGIN_VERSION);
			}
		}

		@Nested
		class TestFailure {

			@Test
			void shouldThrowExceptionIfVersionElementNotFound() {
				when(manifest.getMainAttributes()).thenReturn(attributes);
				doReturn(Collections.enumeration(List.of(manifestUrl))).when(buildProperties).loadManifests();
				doReturn(manifest).when(buildProperties).readManifest(any());

				when(attributes.getValue(anyString())).thenReturn(null);

				assertThrows(ExceptionInInitializerError.class, () -> buildProperties.readPluginVersion());
			}

			@Test
			void shouldThrowExceptionIfManifestNotFound() {
				doReturn(Collections.emptyEnumeration()).when(buildProperties).loadManifests();

				assertThrows(ExceptionInInitializerError.class, () -> buildProperties.readPluginVersion());
			}
		}
	}

	@Nested
	class TestIsOzgPluginManifest {

		@BeforeEach
		void init() {
			when(manifest.getMainAttributes()).thenReturn(attributes);
		}

		@Test
		void shouldCallGetValue() {
			buildProperties.isOzgPluginManifest(manifest);

			verify(attributes).getValue(BuildProperties.KEY_PLUGIN_KEY);
		}

		@Test
		void shouldReturnValue() {
			when(attributes.getValue(anyString())).thenReturn(BuildProperties.OZG_PLUGIN_KEY);

			var result = buildProperties.isOzgPluginManifest(manifest);

			assertThat(result).isTrue();

		}
	}
}