/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.function.Predicate;
import java.util.function.Supplier;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.formcycle.FormCycleConfirmationResponse;
import de.ozgcloud.eingang.formcycle.FormCycleFormData;
import de.ozgcloud.formcycle.attachment.AttachmentMapper;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.ozgcloud.formcycle.attachment.FormcycleAttachmentTestFactory;
import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.ozgcloud.formcycle.errorhandling.OzgPluginSoftError;
import de.ozgcloud.formcycle.errorhandling.Warning;
import de.ozgcloud.formcycle.formdata.FormData;
import de.ozgcloud.formcycle.formdata.FormDataTestFactory;
import de.ozgcloud.formcycle.formdata.OzgCloudFormDataMapper;
import de.ozgcloud.formcycle.formdata.PluginFormDataAdapter;
import de.xima.fc.entities.Attachment;
import lombok.SneakyThrows;

class OzgPluginExecutorTest {

	private final static FormData FORM_DATA = FormDataTestFactory.create();

	@Spy
	@InjectMocks
	private OzgPluginExecutor executor;

	@Mock
	private OzgCloudFormDataMapper ozgCloudFormDataMapper;
	@Mock
	private OzgCloudFormDataHttpClient ozgHttpClient;
	@Mock
	private NodeThrewExceptionFactory exceptionFactory;
	@Mock
	private AttachmentMapper attachmentMapper;
	@Mock
	private PluginFormDataAdapter pluginFormDataAdapter;
	@Mock
	private Supplier<List<Attachment>> attachmentSupplier;
	@Captor
	private ArgumentCaptor<Map<String, Object>> successArgumentCaptor;

	@Nested
	class TestExecution {

		@Mock
		private FormCycleConfirmationResponse response;
		@Mock
		private ExecutionResult executionResult;
		@Mock
		private FormData formDataWithRepresentations;
		@Mock
		private FormData formDataWithAttachments;

		@SneakyThrows
		@BeforeEach
		void setup() {
			when(ozgHttpClient.send(any(), any(), any())).thenReturn(response);
			when(pluginFormDataAdapter.readFormData()).thenReturn(FORM_DATA);
			doReturn(formDataWithRepresentations).when(executor).addRepresentations(any());
			doReturn(formDataWithAttachments).when(executor).addAttachments(any());
			doReturn(executionResult).when(executor).buildExecutionResult(any(), any());
		}

		@Test
		void shouldCallPluginFormDataAdapter() {
			executor.execute();

			verify(pluginFormDataAdapter).readFormData();
		}

		@Test
		void shouldCallAddRepresentations() {
			executor.execute();

			verify(executor).addRepresentations(FORM_DATA);
		}

		@Test
		void shouldCallAddAttachments() {
			executor.execute();

			verify(executor).addAttachments(formDataWithRepresentations);
		}

		@Test
		void shouldCallSendFormData() {
			executor.execute();

			verify(executor).sendFormData(formDataWithAttachments);
		}

		@Test
		void shouldCallBuildExecutionResult() {
			executor.execute();

			verify(executor).buildExecutionResult(response, formDataWithAttachments);
		}

		@Test
		void shouldReturnExecutionResult() {
			var result = executor.execute();

			assertThat(result).isSameAs(executionResult);
		}
	}

	@Nested
	class TestAddRepresentations {

		@Mock
		private Warning warning;

		@Captor
		private ArgumentCaptor<Predicate<String>> filterPredicateCapture;

		private final FormData formData = FormDataTestFactory.createBuilder().clearRepresentations().build();

		@Test
		void shouldCallGetAttachedFiles() {
			doReturn(List.of(FormcycleAttachmentTestFactory.create())).when(executor).getAttachedFiles(any());

			executor.addRepresentations(formData);

			verify(executor).getAttachedFiles(filterPredicateCapture.capture());
			assertThat(filterPredicateCapture.getValue().test(FormDataTestFactory.ATTACHMENT_UUID)).isFalse();
		}

		@Test
		void shouldSetRepresentations() {
			var representation = FormcycleAttachmentTestFactory.create();
			doReturn(List.of(representation)).when(executor).getAttachedFiles(any());

			var result = executor.addRepresentations(formData);

			assertThat(result.getRepresentations()).hasSize(1).first().isEqualTo(representation);
		}

		@Test
		void shouldNotAddAnyWarnings() {
			doReturn(List.of(FormcycleAttachmentTestFactory.create())).when(executor).getAttachedFiles(any());

			var result = executor.addRepresentations(FormDataTestFactory.createBuilder().clearWarnings().build());

			assertThat(result.getWarnings()).isEmpty();
		}

		@Test
		void shouldCallBuildMissingRepresentationsWarning() {
			doReturn(Collections.emptyList()).when(executor).getAttachedFiles(any());
			doReturn(warning).when(executor).buildMissingRepresentationsWarning(any());

			executor.addRepresentations(formData);

			verify(executor).buildMissingRepresentationsWarning(formData.getRequestId());
		}

		@Test
		void shouldAddWhenMissingRepresentations() {
			doReturn(Collections.emptyList()).when(executor).getAttachedFiles(any());
			doReturn(warning).when(executor).buildMissingRepresentationsWarning(any());

			var result = executor.addRepresentations(formData);

			assertThat(result.getWarnings()).hasSize(1).first().isSameAs(warning);
		}

	}

	@Nested
	class TestAddAttachments {

		@Mock
		private Warning warning;

		@Captor
		private ArgumentCaptor<Set<String>> attachmentUuidsCapture;
		@Captor
		private ArgumentCaptor<Predicate<String>> filterPredicateCapture;

		private FormData formData = FormDataTestFactory.createBuilder().clearAttachments().build();

		@Test
		void shouldReturnSameWhenNoAttachmentUuids() {
			var expectedFormData = FormDataTestFactory.createBuilder().clearAttachmentUuids().build();

			var result = executor.addAttachments(expectedFormData);

			assertThat(result).isSameAs(expectedFormData);
		}

		@Test
		void shouldCallGetAttachedFiles() {
			doReturn(List.of(FormcycleAttachmentTestFactory.create())).when(executor).getAttachedFiles(any());

			executor.addAttachments(formData);

			verify(executor).getAttachedFiles(filterPredicateCapture.capture());
			assertThat(filterPredicateCapture.getValue().test(FormDataTestFactory.ATTACHMENT_UUID)).isTrue();
		}

		@Test
		void shouldSetAttachments() {
			var representation = FormcycleAttachmentTestFactory.create();
			doReturn(List.of(representation)).when(executor).getAttachedFiles(any());

			var result = executor.addAttachments(formData);

			assertThat(result.getAttachments()).hasSize(1).first().isEqualTo(representation);
		}

		@Test
		void shouldNotAddAnyWarnings() {
			doReturn(List.of(FormcycleAttachmentTestFactory.create())).when(executor).getAttachedFiles(any());

			var result = executor.addAttachments(FormDataTestFactory.createBuilder().clearWarnings().build());

			assertThat(result.getWarnings()).isEmpty();
		}

		@Test
		void shouldCallBuildMissingAttachmentsWarning() {
			doReturn(Collections.emptyList()).when(executor).getAttachedFiles(any());
			doReturn(warning).when(executor).buildMissingAttachmentsWarning(any());

			executor.addAttachments(formData);

			verify(executor).buildMissingAttachmentsWarning(formData.getRequestId());
		}

		@Test
		void shouldAddWhenMissingAttachments() {
			doReturn(Collections.emptyList()).when(executor).getAttachedFiles(any());
			doReturn(warning).when(executor).buildMissingAttachmentsWarning(any());

			var result = executor.addAttachments(formData);

			assertThat(result.getWarnings()).hasSize(1).first().isSameAs(warning);
		}

	}

	@Nested
	@DisplayName("Attachment's handling")
	class TestGetAttachedFiles {

		@Mock
		private Attachment attachment;
		@Mock
		private Attachment representation;

		@Test
		void shouldReturnEmpty() {
			when(attachmentSupplier.get()).thenReturn(Collections.emptyList());

			var attachedFiles = executor.getAttachedFiles(uuid -> true);

			assertThat(attachedFiles).isEmpty();
		}

		@Test
		void shouldCallMapper() {
			initAttachment();

			executor.getAttachedFiles(uuid -> true);

			verify(attachmentMapper).map(attachment);
		}

		@Test
		void shouldApplyFilterAndReturnAttachments() {
			var expectedAttachment = initAttachment();
			when(representation.getUUID()).thenReturn(UUID.randomUUID().toString());
			when(attachmentSupplier.get()).thenReturn(List.of(representation, attachment));

			var attachedFiles = executor.getAttachedFiles(FormcycleAttachmentTestFactory.UUID::equals);

			assertThat(attachedFiles).hasSize(1).containsExactly(expectedAttachment);
		}

		private FormcycleAttachment initAttachment() {
			var expectedAttachment = FormcycleAttachmentTestFactory.create();
			when(attachmentMapper.map(any())).thenReturn(expectedAttachment);
			when(attachment.getUUID()).thenReturn(FormcycleAttachmentTestFactory.UUID);
			when(attachmentSupplier.get()).thenReturn(List.of(attachment));
			return expectedAttachment;
		}

	}

	@Nested
	class TestBuildMissingRepresentationsWarning {

		private final String REQUEST_ID = "REQUEST_ID";

		@Test
		void shouldSetErrorCode() {
			var warning = executor.buildMissingRepresentationsWarning(REQUEST_ID);

			assertThat(warning.getErrorCode()).isEqualTo(OzgPluginSoftError.MISSING_REPRESENTATION.name());
		}

		@Test
		void shouldSetMessage() {
			var warning = executor.buildMissingRepresentationsWarning(REQUEST_ID);

			assertThat(warning.getMessage()).contains(REQUEST_ID);
		}
	}

	@Nested
	class TestBuildMissingAttachmentsWarning {

		private final String REQUEST_ID = "REQUEST_ID";

		@Test
		void shouldSetErrorCode() {
			var warning = executor.buildMissingAttachmentsWarning(REQUEST_ID);

			assertThat(warning.getErrorCode()).isEqualTo(OzgPluginSoftError.MISSING_ATTACHMENTS.name());
		}

		@Test
		void shouldSetMessage() {
			var warning = executor.buildMissingAttachmentsWarning(REQUEST_ID);

			assertThat(warning.getMessage()).contains(REQUEST_ID);
		}
	}

	@Nested
	@DisplayName("Send data")
	class TestSendData {

		@Mock
		private FormCycleFormData formcycleFormData;
		@Mock
		private FormCycleConfirmationResponse response;

		@Captor
		private ArgumentCaptor<List<FormcycleAttachment>> attachmentsCapture;
		@Captor
		private ArgumentCaptor<List<FormcycleAttachment>> representationsCapture;

		@Test
		void shouldCallMapper() {
			sendFormData();

			verify(ozgCloudFormDataMapper).map(FORM_DATA);
		}

		@Test
		void shouldSendFormData() {
			when(ozgCloudFormDataMapper.map(any())).thenReturn(formcycleFormData);

			sendFormData();

			verify(ozgHttpClient).send(eq(formcycleFormData), any(), any());
		}

		@Test
		void shouldSendAttachments() {
			sendFormData();

			verify(ozgHttpClient).send(any(), attachmentsCapture.capture(), any());
			assertThat(attachmentsCapture.getValue()).hasSize(1).first().extracting("uuid").isEqualTo(FormDataTestFactory.ATTACHMENT_UUID);
		}

		@Test
		void shouldSendRepresentations() {
			sendFormData();

			verify(ozgHttpClient).send(any(), any(), representationsCapture.capture());
			assertThat(representationsCapture.getValue()).hasSize(1).first().extracting("uuid").isEqualTo(FormDataTestFactory.REPRESENTATION_UUID);
		}

		@Test
		void shouldReturnResponse() {
			when(ozgHttpClient.send(any(), any(), any())).thenReturn(response);

			var result = sendFormData();

			assertThat(result).isSameAs(response);
		}

		private FormCycleConfirmationResponse sendFormData() {
			return executor.sendFormData(FORM_DATA);
		}
	}

	@Nested
	class TestBuildExecutionResults {

		@Mock
		private Warning warning;

		@Test
		void shouldSetVorgangNummer() {
			var result = executor.buildExecutionResult(FormCycleConfirmationResponseTestFactory.create(), FORM_DATA);

			assertThat(result.getVorgangnummer()).isEqualTo(FormCycleConfirmationResponseTestFactory.VORGANG_NUMMER);
		}

		@Test
		void shouldSetWarnings() {
			var formData = FormDataTestFactory.createBuilder().clearWarnings().warning(warning).build();

			var result = executor.buildExecutionResult(FormCycleConfirmationResponseTestFactory.create(), formData);

			assertThat(result.getWarnings()).containsExactly(warning);
		}
	}
}