package de.ozgcloud.formcycle.formdata;

import static de.ozgcloud.formcycle.formdata.ItemClass.*;
import static de.ozgcloud.formcycle.formdata.StructureMockFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;

import de.xima.fc.form.common.items.XItem;
import lombok.SneakyThrows;

class FormStructureParserTest {

	private FormStructureParser structureParser;

	@Nested
	class TestParse {

		@Test
		void shouldClearResultNodes() {
			structureParser = new FormStructureParser(Collections.emptyMap());
			injectResultNodes(new ArrayList<>(List.of(FormNodeTestFactory.create())));

			var result = structureParser.parse();

			assertThat(result).isEmpty();
		}

		@Test
		void shouldReturnResultNodes() {
			structureParser = new FormStructureParser(Collections.emptyMap());
			var expectedResultNode = new ArrayList<>(List.of(FormNodeTestFactory.create()));
			injectResultNodes(expectedResultNode);

			var result = structureParser.parse();

			assertThat(result).isSameAs(expectedResultNode);
		}

		@Nested
		class TestHandleInputNode {

			@Test
			void shouldCallHandleInputNode() {
				var items = createItem(TEXT_FIELD);
				structureParser = spy(createParser(items));

				structureParser.parse();

				verify(structureParser).handleInputNode(items.get(ITEM_ID));
			}

			@Test
			void shouldNotCallHandleContainerNode() {
				var items = createItem(TEXT_FIELD);
				structureParser = spy(createParser(items));

				structureParser.parse();

				verify(structureParser, never()).handleContainerNode(any());
			}

		}

		@Nested
		class TestHandleContainerNode {

			@Test
			void shouldHandleContainerNode() {
				var items = createItem(FIELD_SET);
				structureParser = spy(createParser(items));

				structureParser.parse();

				verify(structureParser).handleContainerNode(items.get(ITEM_ID));
			}

			@Test
			void shouldNotCallHandleInputNode() {
				var items = createItem(FIELD_SET);
				structureParser = spy(createParser(items));

				structureParser.parse();

				verify(structureParser, never()).handleInputNode(any());
			}

		}

		private FormStructureParser createParser(Map<String, XItem> items) {
			return new FormStructureParser(items);
		}

	}

	@Nested
	class TestHandleInputNode {

		@BeforeEach
		void init() {
			structureParser = spy(new FormStructureParser(createItem(FIELD_SET)));
		}

		@Test
		void shouldCallGetResultNodes() {
			var xItem = getMock(TEXT_FIELD);

			structureParser.handleInputNode(xItem);

			verify(structureParser).getResultNodes(xItem);
		}

		@Test
		void shouldCallBuildNode() {
			var xItem = getMock(TEXT_FIELD);
			doReturn(new ArrayList<>()).when(structureParser).getResultNodes(any());

			structureParser.handleInputNode(xItem);

			verify(structureParser).buildNode(xItem);
		}

		@Test
		void shouldAddNodeToResultNodes() {
			var resultNodes = new ArrayList<FormNode>();
			doReturn(resultNodes).when(structureParser).getResultNodes(any());
			var formNode = FormNodeTestFactory.create();
			doReturn(formNode).when(structureParser).buildNode(any());

			structureParser.handleInputNode(getMock(TEXT_FIELD));

			assertThat(resultNodes).containsExactly(formNode);
		}
	}

	@Nested
	class TestHandleContainerNode {

		@BeforeEach
		void init() {
			structureParser = spy(new FormStructureParser(createItem(FIELD_SET)));
		}

		@Test
		void shouldCallBuildNode() {
			var xItem = getMock(FIELD_SET);

			structureParser.handleContainerNode(xItem);

			verify(structureParser).buildNode(xItem);
		}

		@Test
		void shouldCallGetResultNodes() {
			var xItem = getMock(FIELD_SET);

			structureParser.handleContainerNode(xItem);

			verify(structureParser).getResultNodes(xItem);
		}

		@Test
		void shouldAddNodeToResultNodes() {
			var resultNodes = new ArrayList<FormNode>();
			doReturn(resultNodes).when(structureParser).getResultNodes(any());
			var formNode = FormNodeTestFactory.create();
			doReturn(formNode).when(structureParser).buildNode(any());

			structureParser.handleContainerNode(getMock(FIELD_SET));

			assertThat(resultNodes).containsExactly(formNode);
		}

		@Test
		void shouldAddNodeToTopOfContainerNodes() {
			var formNode = FormNodeTestFactory.create();
			doReturn(formNode).when(structureParser).buildNode(any());
			var parentNode = injectContainerNodes();
			var fieldSetMock = getMock(FIELD_SET);
			when(fieldSetMock.getParentId()).thenReturn(parentNode.getFirst().getItemId());

			structureParser.handleContainerNode(fieldSetMock);

			assertThat(getContainerNodes()).first().isSameAs(formNode);
		}

		@SuppressWarnings("unchecked")
		@SneakyThrows
		private Deque<FormNode> getContainerNodes() {
			var containerNodesField = FormStructureParser.class.getDeclaredField("containerNodes");
			containerNodesField.setAccessible(true);
			return (Deque<FormNode>) containerNodesField.get(structureParser);
		}
	}

	@Nested
	class TestBuildNode {

		@InjectMocks
		private FormStructureParser structureParser;

		@Test
		void shouldSetId() {
			var formNode = buildNode();

			assertThat(formNode.getItemId()).isEqualTo(ITEM_ID);
		}

		@Test
		void shouldSetTitleInput() {
			var formNode = buildNode();

			assertThat(formNode.getTitle()).isEqualTo(ITEM_TITLE);
		}

		@Test
		void shouldSetTitleFieldSet() {
			var formNode = buildNode(FIELD_SET);

			assertThat(formNode.getTitle()).isEqualTo(ITEM_TITLE);
		}

		@Test
		void shouldSetTitleText() {
			var formNode = buildNode(SPAN);

			assertThat(formNode.getTitle()).isEqualTo(ITEM_TITLE);
		}

		@Test
		void shouldSetName() {
			var formNode = buildNode();

			assertThat(formNode.getName()).isEqualTo(ITEM_NAME);
		}

		@Test
		@DisplayName("should not set title when title property is null")
		void shouldNotSetTitle() {
			var formNode = buildNode(CONTAINER);

			assertThat(formNode.getTitle()).isNull();
		}

		private FormNode buildNode() {
			return buildNode(TEXT_FIELD);
		}

		private FormNode buildNode(ItemClass itemClass) {
			return structureParser.buildNode(getMock(itemClass));
		}
	}

	@Nested
	class TestGetResultNodes {

		private final List<FormNode> expectedResultNodes = List.of(FormNodeTestFactory.create());

		@BeforeEach
		void init() {
			structureParser = spy(new FormStructureParser(createItem(FIELD_SET)));
			injectResultNodes(expectedResultNodes);
		}

		@Test
		void shouldReturnRootNodesWhenNoContainerNodes() {
			var nodes = structureParser.getResultNodes(StructureMockFactory.getMock(TEXT_FIELD));

			assertThat(nodes).isSameAs(expectedResultNodes);
		}

		@Test
		void shouldReturnRootWhenAntragsteller() {
			var antragstellerXItem = getMock(CONTAINER);
			when(antragstellerXItem.getName()).thenReturn(FormStructureParser.CONTAINER_NAME_ANTRAGSTELLER);
			injectContainerNodes(new ArrayDeque<>(List.of(FormNodeTestFactory.createBuilder().itemId(CONTAINER_ITEM_ID).build())));

			var nodes = structureParser.getResultNodes(antragstellerXItem);

			assertThat(nodes).isSameAs(expectedResultNodes);
		}

		@Nested
		class TestGetNextContainer {

			@BeforeEach
			void init() {
				injectContainerNodes();
			}

			@Test
			void shouldReturnRootWhenNoParentNode() {
				var nodes = structureParser.getResultNodes(getMock(CONTAINER));

				assertThat(nodes).isSameAs(expectedResultNodes);
			}

			@Test
			void shouldCallGetNextContainerNode() {
				var parentId = "parentId";
				var xItem = getMock(CONTAINER);
				when(xItem.getParentId()).thenReturn(parentId);

				structureParser.getResultNodes(xItem);

				verify(structureParser).getNextContainerNode(parentId);
			}

			@Test
			void shouldReturnNestedNodes() {
				var formNode = FormNodeTestFactory.create();
				doReturn(Optional.of(formNode)).when(structureParser).getNextContainerNode(anyString());
				var expectedNestedCollection = formNode.getNestedElements();

				var nodes = structureParser.getResultNodes(getMock(CONTAINER));

				assertThat(nodes).isSameAs(expectedNestedCollection);
			}
		}
	}

	@Nested
	class TestGetNextContainerNode {

		@BeforeEach
		void init() {
			structureParser = spy(new FormStructureParser(createItem(FIELD_SET)));
		}

		@Test
		void shouldReturnEmptyWhenNodeStackEmpty() {
			var nextContainerNode = structureParser.getNextContainerNode(ITEM_ID);

			assertThat(nextContainerNode).isEmpty();
		}

		@Test
		void shouldReturnEmpty() {
			injectContainerNodes(new ArrayDeque<>(List.of(FormNodeTestFactory.create())));

			var nextContainerNode = structureParser.getNextContainerNode("parentId");

			assertThat(nextContainerNode).isEmpty();
		}

		@Test
		void shouldReturnParent() {
			var parentId = "parentId";
			FormNode parentNode = FormNodeTestFactory.createBuilder().itemId(parentId).build();
			var parentNodes = injectContainerNodes(new ArrayDeque<>());
			parentNodes.push(parentNode);
			parentNodes.push(FormNodeTestFactory.create());

			var nextContainerNode = structureParser.getNextContainerNode(parentId);

			assertThat(nextContainerNode).contains(parentNode);
		}
	}

	@SneakyThrows
	private List<FormNode> injectResultNodes(List<FormNode> nodes) {
		var containerNodesField = FormStructureParser.class.getDeclaredField("resultNodes");
		containerNodesField.setAccessible(true);
		containerNodesField.set(structureParser, nodes);
		return nodes;
	}

	@SneakyThrows
	private Deque<FormNode> injectContainerNodes() {
		return injectContainerNodes(new ArrayDeque<>(List.of(FormNodeTestFactory.create())));
	}

	@SneakyThrows
	private Deque<FormNode> injectContainerNodes(Deque<FormNode> containerNodes) {
		var containerNodesField = FormStructureParser.class.getDeclaredField("containerNodes");
		containerNodesField.setAccessible(true);
		containerNodesField.set(structureParser, containerNodes);
		return containerNodes;
	}

}